unit LD70909;

interface

uses
  Windows, Classes, ExtCtrls
  ;

const
  POWER_MASK        = $07;
  LEGO_CLOCKWISE    = 1;
  LEGO_COUNTERCW    = 0;

type
  TOutput = 0..7;
  TOutputList = set of TOutput;
  TInput = 0..7;
  TInputList = set of TInput;
  TDirection = (dCW, dCCW);
  TOnOff = (pOn, pOff);
  TTime = 0..255;                       // ? are the larger values relevant ?
  TPower = 0..8;
  TMode = (mSteady, mFlashing);

  TCommError = (ceNoOpen, ceNoConnect, ceSend, ceCheckSum, ceResynch, ceStopped);

  TNotifyCommError = procedure(ErrorCode: TCommError) of object;
  TNotifyInputChange = procedure of object;
  TNotifyConnect = procedure of object;
  TNotifyDisconnect = procedure of object;
  TLoggEvent = procedure(s: string) of object;

  TBoolArray8 = array[0..7] of boolean;
  TIntArray8 = array[0..7] of integer;

  TOutputShadow = record
    OnOff: TOnOff;
    Dir: TDirection;
    Power: TPower;
    Mode: TMode;
  end;

  TLDInterface = class(TObject)
  private
    { Private declarations }
    hCom: THandle;
    fComPort: string;
    fRecMess: string;
    fOldRecMess: string;
    fWatchTime: DWord;
    fAliveTimer: TTimer;
    fRecTimer: TTimer;

    fPressed: TBoolArray8;
    fAngle: TIntArray8;
    fValue: TIntArray8;
    fCount: TIntArray8;

    fConnected: boolean;
    fStopped: boolean;
    fCommError: TCommError;

    fOnCommError: TNotifyCommError;
    fOnInputChange: TNotifyInputChange;
    fOnConnect: TNotifyConnect;
    fOnDisconnect: TNotifyDisconnect;
    fOnLogg: TLoggEvent;

    fOutputShadow: array[0..7] of TOutputShadow;

    procedure AliveTimerTimer(Sender: TObject);
    procedure RecTimerTimer(Sender: TObject);
    procedure CheckInput(port, index: integer);
    procedure CheckStopped;
    procedure AddToLog(s: string);
    procedure ClearInputs;
    procedure SetComPort(const Value: string);
    function GetPressed(Index: integer): boolean;
    function GetAngle(Index: integer): integer;
    function GetValue(Index: integer): integer;
    function GetCount(Index: integer): integer;
    procedure ResetAllShadows;
  public
    { Public declarations }
    constructor Create;
    destructor Destroy; override;

    procedure Connect;
    procedure Disconnect;
    procedure Send(s: string);
    procedure ResetInput(I: integer);

    procedure SetPower(Outputs: TOutputList; Power: TPower);
    procedure TurnOn(Output: TOutput); overload;
    procedure TurnOn(Outputs: TOutputList); overload;
    procedure SetDir(Output: TOutput; Dir: TDirection); overload;
    procedure SetDir(Outputs: TOutputList; Dir: TDirection); overload;
    procedure Reverse(Output: TOutput); overload;
    procedure Reverse(Outputs: TOutputList); overload;
    procedure TurnOff(Output: TOutput); overload;
    procedure TurnOff(Outputs: TOutputList); overload;

    procedure TurnOnDir(Output: TOutput; Dir: TDirection); overload;
    procedure TurnOnDir(Outputs: TOutputList; Dir: TDirection); overload;
    procedure TurnOnDirPower(Output: TOutput; Dir: TDirection; Power: TPower); overload;
    procedure TurnOnDirPower(Outputs: TOutputList; Dir: TDirection; Power: TPower); overload;
    procedure TurnOnOff(Output: TOutput; OnOff: TOnOff); overload;
    procedure TurnOnOff(Outputs: TOutputList; OnOff: TOnOff); overload;
    procedure TurnOnTime(Output: TOutput; Time: TTime);

    procedure SetSteadyMode(Output: TOutput); overload;
    procedure SetSteadyMode(Outputs: TOutputList); overload;
    procedure SetFlashMode(Output: TOutput); overload;
    procedure SetFlashMode(Outputs: TOutputList); overload;
    procedure SetFlashTime(Output: TOutput; OnTime, OffTime: TTime); overload;
    procedure SetFlashTime(Outputs: TOutputList; OnTime, OffTime: TTime); overload;
    procedure SetFlashOnTime(Output: TOutput; OnTime: TTime); overload;
    procedure SetFlashOnTime(Outputs: TOutputList; OnTime: TTime); overload;
    procedure SetFlashOffTime(Output: TOutput; OffTime: TTime); overload;
    procedure SetFlashOffTime(Outputs: TOutputList; OffTime: TTime); overload;

    procedure InterfaceOff;

    property ComPort: string read fComPort write SetComPort;
    property Connected: boolean read fConnected;
    property Pressed[Index: integer]: boolean read GetPressed;
    property Angle[Index: integer]: integer read GetAngle;
    property Count[Index: integer]: integer read GetCount;
    property Value[Index: integer]: integer read GetValue;
    property Stopped: boolean read fStopped;
    property CommError: TCommError read fCommError;

    property OnCommError: TNotifyCommError read fOnCommError write fOnCommError;
    property OnInputChange: TNotifyInputChange read fOnInputChange write fOnInputChange;
    property OnConnect: TNotifyConnect read fOnConnect write fOnConnect;
    property OnDisconnect: TNotifyDisconnect read fOnDisconnect write fOnDisconnect;
    property OnLogg: TLoggEvent read fOnLogg write fOnLogg;
  end;

var
  ErrorTexts        : array[ceNoOpen..ceStopped] of string = (
    'Could not open serial port',
    'No contact with the Lego Brick',
    'Send error',
    'Checksum error',
    'Synch error',
    'Stopped!'
    );

implementation

uses
  SysUtils
  ;

var
  ALIVE             : char = #$02;
  TURN_OFF          : char = #$90;

(*
  TURN_ON           : char = #$91;

  SET_RIGHT         : char = #$93;
  SET_LEFT          : char = #$94;
  REVERSE           : char = #$95;

  SET_POWER         : byte = $B0;
*)

  ReverseDir        : array[dCW..dCCW] of TDirection = (dCCW, dCW);

{ TLDInterface }

constructor TLDInterface.Create;
begin
  fAliveTimer := TTimer.Create(nil);
  fAliveTimer.Enabled := False;
  fAliveTimer.Interval := 1900;
  fAliveTimer.OnTimer := AliveTimerTimer;

  fRecTimer := TTimer.Create(nil);
  fRecTimer.Enabled := False;
  fRecTimer.Interval := 1;
  fRecTimer.OnTimer := RecTimerTimer;

  fComPort := 'COM1';
  fStopped := False;
  fConnected := False;
end;

destructor TLDInterface.Destroy;
begin
  Disconnect;
  fAliveTimer.Free;
  fRecTimer.Free;
end;

{ TLDInterface private functions }

procedure TLDInterface.AddToLog(s: string);
begin
  if Assigned(fOnLogg) then
    fOnLogg(s);
end;

procedure TLDInterface.ClearInputs;
var
  i                 : integer;
begin
  for i := 0 to 7 do begin
    fPressed[i] := False;
    fAngle[i] := 0;
    fValue[i] := 0;
    fCount[i] := 0;
  end
end;

procedure TLDInterface.ResetInput(I: integer);
begin
  fPressed[i] := False;
  fAngle[i] := 0;
  fValue[i] := 0;
  fCount[i] := 0;
end;

procedure TLDInterface.CheckInput(Port, Index: integer);
var
  TouchEvent        : boolean;
  NewValue          : DWord;
  Direction         : integer;
  TurnAmount        : integer;
begin
 // Touch info
  TouchEvent := FALSE;
  if (Ord(fRecMess[Index + 1]) = $2E) then begin
    TouchEvent := TRUE;
    if not fPressed[Port] then begin
      fPressed[Port] := True;
      // Room for notification
    end
  end
  else begin
    if fPressed[Port] then begin
      fPressed[Port] := False;
      TouchEvent := TRUE;
      // Room for notification
    end
  end;

 // Light and Temperature info (also resistance)
  NewValue := Ord(fRecMess[Index + 1]) * 4 + (Ord(fRecMess[Index + 2]) and $C0) div 64;
  if (fValue[Port] <> NewValue) then begin
    fValue[Port] := NewValue;
    // Room for notification
  end;

(*
  // Status
  fStatus[Port] := Ord(RecMess[Index + 2]) and $3F;
*)

 // Angle Info
  if (Ord(fRecMess[Index + 2]) and $04) <> 0 then
    Direction := LEGO_CLOCKWISE
  else
    Direction := LEGO_COUNTERCW;

  TurnAmount := (Ord(fRecMess[Index + 2]) and $03);
  if TurnAmount in [$01, $02, $03] then begin
    if (Direction = LEGO_CLOCKWISE) then begin
      fAngle[Port] := (fAngle[Port] + TurnAmount) mod 16;
      fCount[Port] := (fCount[Port] + TurnAmount);
    end
    else begin
      fAngle[Port] := (16 + fAngle[Port] - TurnAmount) mod 16;
      fCount[Port] := (fCount[Port] - TurnAmount);
    end;
    // Room for notification
  end

end;

procedure TLDInterface.CheckStopped;
var
  Result            : boolean;
begin
  if (Ord(fRecMess[1]) and $10) <> 0 then begin
    // Stop button pressed on interface
    Result := True;
  end
  else begin
    Result := False;
  end;
  if Result <> fStopped then begin
    fStopped := Result;
    if Assigned(fOnCommError) then fOnCommError(ceStopped);
  end
end;

procedure TLDInterface.AliveTimerTimer(Sender: TObject);
var
  NumBytes          : DWord;
begin
  WriteFile(hCom, ALIVE, 1, NumBytes, nil)
end;

procedure TLDInterface.RecTimerTimer(Sender: TObject);
var
  Buffer            : string;
  s                 : string;
  NumBytes          : DWord;
  i, j              : integer;
  CheckSum          : integer;
begin
  SetLength(Buffer, 256);
  ReadFile(hCom, Buffer[1], 19, NumBytes, nil);
  if NumBytes > 0 then begin
    fWatchTime := GetTickCount;
    i := 1;
    while NumBytes > 0 do begin
      fRecMess := fRecMess + Buffer[i];
      if Length(fRecMess) = 19 then begin
        if fOldRecMess <> fRecMess then begin
          if (fRecMess[1] in [#$00, #$10]) and (fRecMess[2] = #$00) then begin
            s := '';
            CheckSum := 0;
            for j := 1 to 19 do begin
              s := s + Format('%x ', [Ord(fRecMess[j])]);
              CheckSum := (CheckSum + Ord(fRecMess[j])) and $FF;
            end;
            if CheckSum <> $FF then begin
              AddToLog('Checksum error!');
              if Assigned(fOnCommError) then fOnCommError(ceChecksum);
            end
            else begin
              AddToLog(s);
              CheckStopped;
              CheckInput(3, 2);
              CheckInput(7, 4);
              CheckInput(2, 6);
              CheckInput(6, 8);
              CheckInput(1, 10);
              CheckInput(5, 12);
              CheckInput(0, 14);
              CheckInput(4, 16);
              if Assigned(fOnInputChange) then fOnInputChange;
              fOldRecMess := fRecMess;
            end;
            fRecMess := '';
          end
          else begin                    // Out of synch??
            Delete(fRecMess, 1, 1);
            if Assigned(fOnCommError) then fOnCommError(ceResynch);
          end
        end
        else begin
          fRecMess := '';
        end
      end;
      Inc(i);
      Dec(NumBytes);
    end
  end;

  if (GetTickCount - fWatchTime) > 1000 then begin
    AddToLog('Lost contact with the Lego Control Box!');
    Disconnect;
    fRecTimer.Enabled := False;
    if Assigned(fOnCommError) then fOnCommError(ceNoConnect);
  end
end;

{ TLDInterface public functions }

procedure TLDInterface.Connect;
var
  dwError           : DWord;
  NumBytes          : DWord;
  Current           : integer;
  Count             : integer;
  Found             : boolean;
  fResult           : longbool;
  dcb               : TDCB;
  Buffer            : string;
  Info              : string;
  tmo               : TCommTimeouts;
  Cmd               : string;
  tt                : DWord;
  Tries             : integer;
begin
  fConnected := False;
 // Open serial
  hCom := CreateFile(PChar(fComPort),
    GENERIC_READ or GENERIC_WRITE,
    0,                                  // comm devices must be opened w/exclusive-access
    nil,                                // no security attrs
    OPEN_EXISTING,                      // comm devices must use OPEN_EXISTING
    0,
    0                                   // hTemplate must be NULL for comm devices
    );

  if (hCom = INVALID_HANDLE_VALUE) then begin
    dwError := GetLastError();
    AddToLog('Can''t create serial port handle');
    if Assigned(fOnCommError) then fOnCommError(ceNoOpen);
    exit;
  end;

  fResult := GetCommState(hCom, dcb);
  if (not fResult) then begin
    AddToLog('Can''t get initial com port parameters');
    if Assigned(fOnCommError) then fOnCommError(ceNoOpen);
    exit;
  end;

  dcb.BaudRate := 9600;
  dcb.ByteSize := 8;
  dcb.Parity := NOPARITY;
  dcb.StopBits := ONESTOPBIT;

 // Reset the com port to the new configuration
  fResult := SetCommState(hCom, dcb);
  if (not fResult) then begin
    AddToLog('Can''t set communication parameters');
    if Assigned(fOnCommError) then fOnCommError(ceNoOpen);
    exit;
  end;
  AddToLog('COM2 successfully opened...');

 // Set communications timeout values
 // New setup reads all available chars, may return zero

  tmo.ReadIntervalTimeout := MAXDWORD;  // 0
  tmo.ReadTotalTimeoutMultiplier := 0;  //500;
  tmo.ReadTotalTimeoutConstant := 0;    //1000;
  tmo.WriteTotalTimeoutMultiplier := 0;
  tmo.WriteTotalTimeoutConstant := 0;
  fResult := fResult and SetCommTimeouts(hCom, tmo);

  SetLength(Info, 256);

  Tries := 5;
  repeat
    tt := GetTickCount;
 // Send greeting to lego box
    buffer := 'p'#0'###Do you byte, when I knock?$$$';
    WriteFile(hCom, Buffer[1], 34, numbytes, nil);
    if (numbytes <> 34) then begin
      AddToLog('Write failed');
      if Assigned(fOnCommError) then fOnCommError(ceSend);
    end;
 // Read in ###Just a bit off the block!$$$
 // Not much checking, just waiting for the $$$
    Current := 1;
    Count := 0;
    Found := FALSE;

 // Not very smart code yet, but seems to work...
    while (not Found) and ((GetTickCount - tt) < 200) do begin
      if ReadFile(hCom, Info[Current], 1, NumBytes, nil) and (NumBytes <> 0) then begin
        if (Info[Current] = '$') then
          Inc(Count);
        if (Count = 3) then
          Found := TRUE;
        Inc(Current, NumBytes);
      end
    end;
    Dec(Tries);
  until Found or (Tries <= 0);

  if not Found then begin
    AddToLog('Lego Control Box is not responding.');
    if Assigned(fOnCommError) then fOnCommError(ceNoConnect);
    CloseHandle(hCom);
    exit;
  end;

  fConnected := True;
  if Assigned(fOnConnect) then fOnConnect;
  AddToLog(Copy(info, Current - 32, 32));
  fAliveTimer.Enabled := True;
  fRecTimer.Enabled := True;
  fWatchTime := GetTickCount;
  fRecMess := '';
  fOldRecMess := '';
end;

procedure TLDInterface.Disconnect;
var
  Cmd               : string;
  NumBytes          : DWord;
begin
  if fConnected then begin
    fAliveTimer.Enabled := False;
    Cmd := TURN_OFF + #$FF;
    WriteFile(hCom, Cmd[1], Length(Cmd), NumBytes, nil);
    CloseHandle(hCom);
    fRecTimer.Enabled := False;
    fConnected := False;
  end;
  if Assigned(fOnDisconnect) then fOnDisconnect;
end;

procedure TLDInterface.Send(s: string);
var
  BytesSent         : DWord;
begin
  if fConnected then begin
    WriteFile(hCom, s[1], Length(s), BytesSent, nil);
    if BytesSent <> Length(s) then
      if Assigned(fOnCommError) then fOnCommError(ceSend);
  end
end;

{ TLDInterface property setters/getters }

procedure TLDInterface.SetComPort(const Value: string);
var
  WasConnected      : boolean;
begin
  if fComPort <> Value then begin
    WasConnected := fConnected;
    if WasConnected then DisConnect;
    fComPort := Value;
    if WasConnected then Connect;
  end
end;

function TLDInterface.GetPressed(Index: integer): boolean;
begin
  Result := fPressed[Index];
end;

function TLDInterface.GetAngle(Index: integer): integer;
begin
  Result := fAngle[Index];
end;

function TLDInterface.GetCount(Index: integer): integer;
begin
  Result := fCount[Index];
end;

function TLDInterface.GetValue(Index: integer): integer;
begin
  Result := fValue[Index];
end;

{ Output commands }

procedure TLDInterface.ResetAllShadows;
var
  i                 : TOutput;
begin
  for i := 0 to 7 do begin
    fOutputShadow[i].OnOff := pOff;
    fOutputShadow[i].Dir := dCCW;
    fOutputShadow[i].Power := 8;
    fOutputShadow[i].Mode := mSteady;
  end
end;

procedure TLDInterface.InterfaceOff;
begin
  Send(Chr($70));
  ResetAllShadows;
end;

procedure TLDInterface.Reverse(Output: TOutput);
begin
  Send(Chr($20 or Output));
  fOutputShadow[Output].Dir := ReverseDir[fOutputShadow[Output].Dir];
end;

procedure TLDInterface.Reverse(Outputs: TOutputList);
var
  i                 : TOutput;
begin
  Send(Chr($95) + Chr(byte(Outputs)));
  for i := 0 to 7 do
    if i in Outputs then
      fOutputShadow[i].Dir := ReverseDir[fOutputShadow[i].Dir];
end;

procedure TLDInterface.SetDir(Output: TOutput; Dir: TDirection);
begin
  case Dir of
    dCW: Send(Chr($48 or Output));
    dCCW: Send(Chr($40 or Output));
  end;
  fOutputShadow[Output].Dir := Dir;
end;

procedure TLDInterface.SetDir(Outputs: TOutputList; Dir: TDirection);
var
  i                 : TOutput;
begin
  case Dir of
    dCW: Send(Chr($94) + Chr(byte(Outputs)));
    dCCW: Send(Chr($93) + Chr(byte(Outputs)));
  end;
  for i := 0 to 7 do
    if i in Outputs then
      fOutputShadow[i].Dir := Dir;
end;

procedure TLDInterface.SetFlashMode(Output: TOutput);
begin
  SetFlashMode([Output]);
end;

procedure TLDInterface.SetFlashMode(Outputs: TOutputList);
var
  i                 : TOutput;
begin
  Send(Chr($81) + Chr(byte(Outputs)));
  for i := 0 to 7 do
    if i in Outputs then
      fOutputShadow[i].Mode := mFlashing;
end;

procedure TLDInterface.SetFlashOffTime(Output: TOutput; OffTime: TTime);
begin
  Send(Chr($E0 or Output) + Chr(OffTime));
end;

procedure TLDInterface.SetFlashOffTime(Outputs: TOutputList; OffTime: TTime);
var
  i                 : TOutput;
begin
  for i := 0 to 7 do
    if i in Outputs then
      SetFlashOffTime(i, OffTime);
end;

procedure TLDInterface.SetFlashOnTime(Output: TOutput; OnTime: TTime);
begin
  Send(Chr($E8 or Output) + Chr(OnTime));
end;

procedure TLDInterface.SetFlashOnTime(Outputs: TOutputList; OnTime: TTime);
var
  i                 : TOutput;
begin
  for i := 0 to 7 do
    if i in Outputs then
      SetFlashOnTime(i, OnTime);
end;

procedure TLDInterface.SetFlashTime(Output: TOutput; OnTime, OffTime: TTime);
begin
  SetFlashOnTime(Output, OnTime);
  SetFlashOffTime(Output, OffTime);
end;

procedure TLDInterface.SetFlashTime(Outputs: TOutputList; OnTime, OffTime: TTime);
var
  i                 : TOutput;
begin
  for i := 0 to 7 do
    if i in Outputs then begin
      SetFlashOnTime(i, OnTime);
      SetFlashOffTime(i, OffTime);
    end
end;

procedure TLDInterface.SetPower(Outputs: TOutputList; Power: TPower);
var
  i                 : TOutput;
begin
  if Power = 0 then begin
    TurnOff(Outputs);
  end
  else begin
    Send(Chr($B0 or (Power - 1)) + Chr(byte(Outputs)));
    for i := 0 to 7 do
      if i in Outputs then begin
        fOutputShadow[i].Power := Power;
      end
  end
end;

procedure TLDInterface.SetSteadyMode(Output: TOutput);
begin
  SetSteadyMode([Output]);
end;

procedure TLDInterface.SetSteadyMode(Outputs: TOutputList);
var
  i                 : TOutput;
begin
  Send(Chr($80) + Chr(byte(Outputs)));  // Turn off and set steady
  for i := 0 to 7 do
    if (i in Outputs) and (fOutputShadow[i].OnOff = pOn) then
      TurnOn(i);
end;

procedure TLDInterface.TurnOff(Output: TOutput);
begin
  Send(Chr($30 or Output));
  fOutputShadow[Output].OnOff := pOff;
end;

procedure TLDInterface.TurnOff(Outputs: TOutputList);
var
  i                 : TOutput;
begin
  Send(Chr($90) + Chr(byte(Outputs)));
  for i := 0 to 7 do
    if i in Outputs then
      fOutputShadow[i].OnOff := pOff;
end;

procedure TLDInterface.TurnOn(Output: TOutput);
begin
  Send(Chr($28 or Output));
  fOutputShadow[Output].OnOff := pOn;
end;

procedure TLDInterface.TurnOn(Outputs: TOutputList);
var
  i                 : TOutput;
begin
  Send(Chr($91) + Chr(byte(Outputs)));
  for i := 0 to 7 do
    if i in Outputs then
      fOutputShadow[i].OnOff := pOn;
end;

procedure TLDInterface.TurnOnDir(Output: TOutput; Dir: TDirection);
begin
  SetDir(Output, Dir);
  TurnOn(Output);
end;

procedure TLDInterface.TurnOnDir(Outputs: TOutputList; Dir: TDirection);
begin
  SetDir(Outputs, Dir);
  TurnOn(Outputs);
end;

procedure TLDInterface.TurnOnDirPower(Output: TOutput; Dir: TDirection; Power: TPower);
begin
  SetDir(Output, Dir);
  SetPower([Output], Power);
  TurnOn(Output);
end;

procedure TLDInterface.TurnOnDirPower(Outputs: TOutputList; Dir: TDirection; Power: TPower);
begin
  SetDir(Outputs, Dir);
  SetPower(Outputs, Power);
  TurnOn(Outputs);
end;

procedure TLDInterface.TurnOnOff(Output: TOutput; OnOff: TOnOff);
begin
  case OnOff of
    pOn: TurnOn(Output);
    pOff: TurnOff(Output);
  end
end;

procedure TLDInterface.TurnOnOff(Outputs: TOutputList; OnOff: TOnOff);
begin
  case OnOff of
    pOn: TurnOn(Outputs);
    pOff: TurnOff(Outputs);
  end
end;

procedure TLDInterface.TurnOnTime(Output: TOutput; Time: TTime);
begin
  Send(Chr($C0 or Output) + Chr(Time));
  fOutputShadow[Output].OnOff := pOff;  // Count this as a TurnOff
end;

end.

